//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "Group.hpp"
//------------------------------------------------------------------------------
#include <iostream>
//------------------------------------------------------------------------------
#include "MyException.hpp"
#include "Screen.hpp"
#include "AllObj.hpp"

#include "glUtils.hpp"

#include <luabind/nil.hpp>
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
bool compareZpos(const Drawable *d1, const Drawable *d2)
{
    return (d1->GetZpos() < d2->GetZpos());
}
//------------------------------------------------------------------------------
Group::Group(Group *parent, const std::string &name) : Drawable(parent, name), m_scale(Vector2D(1.f, 1.f)), m_cutRect(sf::FloatRect(0, 0, 0, 0)), m_isCut(false)
{
    //ctor
}
//------------------------------------------------------------------------------
Group::~Group()
{
    Clear();
}
//------------------------------------------------------------------------------
void Group::DestroyChildren()
{
    // We need to notice the children
    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            it2->second->Destroy();
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
    {
        (*it)->Destroy();
    }
}
//------------------------------------------------------------------------------
int Group::GetChildCount() const
{
    return GetPhysicChildCount() + GetDrawableChildCount();
}
//------------------------------------------------------------------------------
/** Return the number of physic children that this group contains */
int Group::GetPhysicChildCount() const
{
    return m_phTbl.size();
}
//------------------------------------------------------------------------------
/** Return the number of drawable children that this group contains */
int Group::GetDrawableChildCount() const
{
    int count = 0;

    for(DrawZPosMap::const_iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        count += it->second.size();
    }

    return count;
}
//------------------------------------------------------------------------------
/** Return the physic child number id */
PhysicObject *Group::GetPhysicChild(int id)
{
    if(id < 0 || id >= GetPhysicChildCount())
    {
        MY_THROW("PhysicObject::GetPhysicChild() : out of range !")
    }

    return m_phTbl[id];
}
//------------------------------------------------------------------------------
/** Return the drawable child number id */
Drawable *Group::GetDrawableChild(int id)
{
    if(id < 0)
    {
        MY_THROW("PhysicObject::GetDrawableChild() : out of range !")
    }

    size_t pos = 0;
    for(DrawZPosMap::const_iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        const size_t size = it->second.size();

        if(size_t(id) < pos + size)
        {
            int currentPos = pos;

            for(DrawLowZPosMap::const_iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
            {
                if(currentPos == id)
                {
                    return it2->second;
                }
                currentPos++;
            }
        }

        pos += size;
    }

    MY_THROW("PhysicObject::GetDrawableChild() : out of range !")
}
//------------------------------------------------------------------------------
void Group::RegisterClone(Group *newParent)
{
    // We change the parent if it's needed
    if(newParent != NULL)
    {
        m_parent = newParent;
    }

    // /!\ Specific to Drawable object : we need to refresh the lowPos
    m_lowZpos = GetParent()->Add(this);

    // We notice the object that the parent has changed => Will attach the objet to the pysic world
    // !! Already done in constructor !
    //OnChangeScreen(NULL, GetParentScreen());
}
//------------------------------------------------------------------------------
Group::Group(const Group &grp, Group *newParent) : Drawable(grp)
{
    if(newParent != NULL)
    {
        m_parent = newParent; // We need to set the new parent at first in the group so that physic objects won't register in any other group.
    }

    // Do not copy m_eventToBeDestroyedOnDelete !

    m_scale = grp.m_scale;
    m_isCut = grp.m_isCut;
    m_cutRect = grp.m_cutRect;

    m_drawTbl.clear();

    for(DrawZPosMap::const_iterator it = grp.m_drawTbl.begin(); it != grp.m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::const_iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            it2->second->CloneToGroup(this);
        }
    }

    m_phTbl.clear();
    // iterators are invalidated for vectors
    for(std::vector<PhysicObject*>::const_iterator it = grp.m_phTbl.begin(); it != grp.m_phTbl.end(); it++)
    {
        (*it)->CloneToGroup(this);
    }
}
//------------------------------------------------------------------------------
void Group::OnChangeScreen(Screen *backScreen, Screen *newScreen)
{
    if(backScreen != newScreen/* && backScreen != NULL && newScreen != NULL*/) // If we don't change screen, this is unusefull
    {
        // We need to notice the children

        for(DrawZPosMap::const_iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
        {
            for(DrawLowZPosMap::const_iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
            {
                it2->second->OnChangeScreen(backScreen, newScreen);
            }
        }

        for(std::vector<PhysicObject*>::const_iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
        {
            (*it)->OnChangeScreen(backScreen, newScreen);
        }
    }
}
//------------------------------------------------------------------------------
// FindByName
#define X(typeName, typeId) \
        typeName *Group::FindByName ## typeName(const std::string &name) \
        { \
            Object *obj = FindByNameUncast(name); \
            if(obj->GetType() == typeId) \
            { \
                return static_cast<typeName*>(obj); \
            } \
            else \
            { \
                MY_THROW("Group::FindByName"# typeName "() : can't cast '" << name << "', bad type !"); \
            } \
        }
    X_OBJ_LIST
#undef X
//------------------------------------------------------------------------------
Group *Group::CreateGroup(const std::string &name)
{
    return new Group(this, name);
}
//------------------------------------------------------------------------------
ParticleSource *Group::CreateParticleSource(const std::string &name, const ParticleModel &model, int maxParticleNumber)
{
    return new ParticleSource(this, name, model, maxParticleNumber);
}
//------------------------------------------------------------------------------
Shape *Group::CreateShape(const std::string &name)
{
    return new Shape(this, name);
}
//------------------------------------------------------------------------------
Sprite *Group::CreateSprite(const std::string &name)
{
    return new Sprite(this, name);
}
//------------------------------------------------------------------------------
Text *Group::CreateText(const std::string &name, const std::string &text)
{
    return new Text(this, name, text);
}
//------------------------------------------------------------------------------
PostFX *Group::CreatePostFX(const std::string &name)
{
    return new PostFX(this, name);
}
//------------------------------------------------------------------------------
ScreenRenderer *Group::CreateScreenRenderer(const std::string &name)
{
    return new ScreenRenderer(this, name);
}
//------------------------------------------------------------------------------
PhLine *Group::CreatePhLine(int phGroupId, const std::string &name, const Vector2D &pt1, const Vector2D &pt2)
{
    return new PhLine(this, phGroupId, name, pt1, pt2);
}
//------------------------------------------------------------------------------
PhRect *Group::CreatePhRect(int phGroupId, const std::string &name, const Vector2D &size)
{
    return new PhRect(this, phGroupId, name, size);
}
//------------------------------------------------------------------------------
PhTriangle *Group::CreatePhTriangle(int phGroupId, const std::string &name, const Vector2D &pt1, const Vector2D &pt2, const Vector2D &pt3)
{
    return new PhTriangle(this, phGroupId, name, pt1, pt2, pt3);
}
//------------------------------------------------------------------------------
PhShape *Group::CreatePhShape(int phGroupId, const std::string &name)
{
    return new PhShape(this, phGroupId, name);
}
//------------------------------------------------------------------------------
PhCircle *Group::CreatePhCircle(int phGroupId, const std::string &name, const Vector2D &center, float radius)
{
    return new PhCircle(this, phGroupId, name, center, radius);
}
//------------------------------------------------------------------------------
void Group::SetScale(const Vector2D &scale)
{
    m_scale = scale;
}
//------------------------------------------------------------------------------
Vector2D Group::GetScale()
{
    return m_scale;
}
//------------------------------------------------------------------------------
void Group::CleanDeleted()
{
    // Delete objects regitered in this group

    // Clean childs
    for(DrawZPosMap::const_iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::const_iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            it2->second->CleanDeleted();
        }
    }

    for(std::vector<PhysicObject*>::const_iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
    {
        (*it)->CleanDeleted();
    }


    for(std::vector<Object*>::iterator it = m_toDelete.begin(); it != m_toDelete.end(); it++ )
    {
        (*it)->ChangeParent(NULL);
        delete *it;
    }
    m_toDelete.clear();
}
//------------------------------------------------------------------------------
void Group::AddToDelete(Object *obj)
{
    if(find(m_toDelete.begin(), m_toDelete.end(), obj) == m_toDelete.end())
    {
        m_toDelete.push_back(obj);
    }
    else
    {
        cerr << "AddToDelete: Object already in delete list (object name: '" << obj->GetName() << "') !" << endl;
    }
}
//------------------------------------------------------------------------------
void Group::Clear()
{
    // Events
    for(std::vector<EventHandlerBase*>::iterator it = m_eventToBeDestroyedOnDelete.begin(); it != m_eventToBeDestroyedOnDelete.end(); it++ )
    {
        (*it)->Destroy();
    }
    m_eventToBeDestroyedOnDelete.clear();

    // Drawable
    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            delete it2->second;
        }
        it->second.clear();
    }
    m_drawTbl.clear();

    // Physic objects
    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
    {
        delete *it;
    }
    m_phTbl.clear();
}
//------------------------------------------------------------------------------
void Group::Render(sf::RenderTarget &App)
{
    const sf::View &backView = App.GetView();
    sf::View tmpView;

    // Scale
    if(m_scale != Vector2D(1.f, 1.f))
    {
        sf::FloatRect backRect = backView.GetRect();
        Vector2D globalPos = GetGlobalPosition();

        sf::FloatRect viewRect;
        viewRect.Left = (globalPos.x - backRect.Left) * ((m_scale.x - 1.f) / m_scale.x) + backRect.Left;
        viewRect.Top = (globalPos.y - backRect.Top) * ((m_scale.y - 1.f) / m_scale.y) + backRect.Top;
        viewRect.Bottom = backRect.GetHeight() / m_scale.x + viewRect.Top;
        viewRect.Right = backRect.GetWidth() / m_scale.y + viewRect.Left;
        tmpView.SetFromRect(viewRect);
        App.SetView(tmpView);
    }

    // Draw
    if(IsVisible())
    {
        // Cut
        if(m_isCut)
        {
            Vector2D globalPos = GetGlobalPosition();
            EnableOpenGLCut(sf::FloatRect(m_cutRect.Left + globalPos.x, m_cutRect.Top + globalPos.y, m_cutRect.Right + globalPos.x, m_cutRect.Bottom + globalPos.y), App);
        }

        // Render all drawables
        for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
        {
            for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
            {
                it2->second->Render(App);
            }
        }

        // Cut
        if(m_isCut)
        {
            DisableOpenGLCut();
        }
    }

    // Debug
    if(Global::IsShowPhysicObjects())
    {
        // Group center
        Vector2D pos = GetGlobalPosition();
        sf::Shape circle = sf::Shape::Circle(pos.x, pos.y, 4, Color(0, 255, 0));
        App.Draw(circle);

        // Physic objects
        for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
        {
            (*it)->RenderShape(App);
        }
    }

    App.SetView(backView);
}
//------------------------------------------------------------------------------
int Group::Add(Drawable *obj)
{
    int lowZPos = 0;
    DrawLowZPosMap &map = m_drawTbl[obj->GetZpos()];
    if(!map.empty())
    {
        DrawLowZPosMap::iterator last = map.end();
        last --; // We use the low id of the last element
        lowZPos = last->first + 1; // The new lowZPos is the last lowzpos of the list + 1
    }

    map[lowZPos] = obj;
    return lowZPos;
}
//------------------------------------------------------------------------------
void Group::Add(PhysicObject *obj)
{
    m_phTbl.push_back(obj);
}
//------------------------------------------------------------------------------
void Group::Remove(Drawable *obj)
{
    bool found = false;

    DrawLowZPosMap &map = m_drawTbl[obj->GetZpos()];
    DrawLowZPosMap::iterator it = map.find(obj->GetLowZPos());
    if(it->second == obj)
    {
        map.erase(it);
        found = true;
    }

    if(!found)
    {
        MY_THROW("Group::Remove : Pointer of object '" << obj->GetName() << "' not found !");
    }
}
//------------------------------------------------------------------------------
void Group::Remove(PhysicObject *obj)
{
    std::vector<PhysicObject*>::iterator it = find(m_phTbl.begin(), m_phTbl.end(), obj);
    if(it != m_phTbl.end())
    {
        m_phTbl.erase(it);
    }
    else
    {
        MY_THROW("Group::Remove : Pointer not found !");
    }
}
//------------------------------------------------------------------------------
void Group::ProceedMvt(float time)
{
    Drawable::ProceedMvt(time);

    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            it2->second->ProceedMvt(time);
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
    {
        (*it)->ProceedMvt(time);
    }
}
//------------------------------------------------------------------------------
/*void Group::SetPosition(const Vector2D &pos)
{
    Object::SetPosition(pos);
    m_sfDrawable->SetPosition(GetPosition());
}*/
//------------------------------------------------------------------------------
Shape *Group::CreateShapeLine(const std::string &name, const Vector2D& P1, const Vector2D& P2, float Thickness, const Color& Col, float Outline, const Color& OutlineCol)
{
	Shape *shape = CreateShape(name);
	shape->MakeLine(P1, P2, Thickness, Col, Outline, OutlineCol);
	return shape;
}
//------------------------------------------------------------------------------
Shape *Group::CreateShapeRectangle(const std::string &name, const Vector2D& P1, const Vector2D& P2, const Color& Col, float Outline, const Color& OutlineCol)
{
	Shape *shape = CreateShape(name);
	shape->MakeRectangle(P1, P2, Col, Outline, OutlineCol);
	return shape;
}
//------------------------------------------------------------------------------
Shape *Group::CreateShapeCircle(const std::string &name, const Vector2D& Center, float Radius, const Color& Col, float Outline, const Color& OutlineCol)
{
	Shape *shape = CreateShape(name);
	shape->MakeCircle(Center, Radius, Col, Outline, OutlineCol);
	return shape;
}
//------------------------------------------------------------------------------
Object *Group::FindByNameUncast(const std::string &name, const size_t &id)
{
    size_t current = 0;

    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            if(it2->second->GetName() == name)
            {
                if(id == current)
                {
                    return it2->second;
                }

                current++;
            }
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++)
    {
        if((*it)->GetName() == name)
        {
            if(id == current)
            {
                return *it;
            }

            current++;
        }
    }

    MY_THROW("Group::FindByName() : Object '" << name << "' not found in group '" << GetName() << "'!");
}
//------------------------------------------------------------------------------
/** Get a lua table containing all its childs.\\n
  * Warning: no particular order are warranted.
  */
luabind::object Group::GetChildTable()
{
    luabind::object table = luabind::newtable(Global::GetLuaState());

    size_t current = 1; // Index start at 1 in lua !

    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            table[current] = CastObject(it2->second);
            current++;
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++)
    {
        table[current] = CastObject(*it);
        current++;
    }

    return table;
}
//------------------------------------------------------------------------------

/** Get a child from its index in the group (a number for 1 to GetChildCount().\\n
  * Warning: no particular order are warranted.
  */
luabind::object Group::FindByIndex(int id)
{
    Object *obj = NULL;
    int current = 1; // Index start at 1 in lua !

    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end() && obj == NULL; it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            if(id == current)
            {
                obj = it2->second;
                break;
            }

            current++;
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end() && obj == NULL; it++)
    {
        if(id == current)
        {
            obj = *it;
            break;
        }

        current++;
    }

    if(obj == NULL)
    {
        MY_THROW("Group::FindByIndex() : Object not found in group '" << GetName() << "' (id is too high) !");
    }

    // Cast
    return CastObject(obj);
}
//------------------------------------------------------------------------------
luabind::object Group::CastObject(Object *obj)
{
    if(obj == NULL)
    {
        MY_THROW("Group::CastObject() : Obj is NULL in group '" << GetName() << "'!");
    }

    luabind::object luaObj;
    switch(obj->GetType())
    {
        #define X(typeName, typeId) \
                case typeId: \
                    luaObj = luabind::object(Global::GetLuaState(), static_cast<typeName*>(obj)); \
                    break;
            X_OBJ_LIST
        #undef X
        default:
            MY_THROW("Group::CastObject() : Unknown type << '" << obj->GetType() << "' in group '" << GetName() << "'!");
            break;
    }

    return luaObj;
}
//------------------------------------------------------------------------------
luabind::object Group::FindByName(const std::string &name)
{
    Object *obj = NULL;

    try
    {
        obj = FindByNameUncast(name);
    }
    catch(const std::exception &e)
    {
        return luabind::object(); // nil
    }

    return CastObject(obj);
}
//------------------------------------------------------------------------------

void Group::PropagateRotation(float deltaRotation)
{
    /*for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            if(it2->second->GetType() == OBJ_TYPE_DRAWABLE_GROUP)
            {
                (static_cast<Group*>(it2->second))->PropageRotation(deltaRotation);
            }

            it2->second->SetPosition(PointRotate(Vector2D(0, 0), it2->second->GetPosition(), deltaRotation));
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
    {
        (*it)->SetPosition(PointRotate(Vector2D(0, 0), (*it)->GetPosition(), deltaRotation));
    }*/

    for(DrawZPosMap::iterator it = m_drawTbl.begin(); it != m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            float dist, angle;
            it2->second->GetPolarPosition(dist, angle);
            it2->second->SetPolarPosition(dist, angle + deltaRotation);

            if(it2->second->GetType() == OBJ_TYPE_DRAWABLE_GROUP)
            {
                (static_cast<Group*>(it2->second))->PropagateRotation(deltaRotation);
            }
        }
    }

    for(std::vector<PhysicObject*>::iterator it = m_phTbl.begin(); it != m_phTbl.end(); it++ )
    {
        float dist, angle;
        (*it)->GetPolarPosition(dist, angle);
        (*it)->SetPolarPosition(dist, angle + deltaRotation);
    }
}

//------------------------------------------------------------------------------
void Group::SetRotation(float rotation)
{
    float backRotation = GetRotation();
    Object::SetRotation(rotation);

    PropagateRotation(rotation - backRotation);
}
//------------------------------------------------------------------------------
/** Set the cut rectangle. You should enable cutting. */
void Group::SetCutRect(const Vector2D &position, const Vector2D &size)
{
    m_cutRect.Left = position.x;
    m_cutRect.Top = position.y;

    m_cutRect.Right = position.x + max(0.f, size.x);
    m_cutRect.Bottom = position.y + max(0.f, size.y);
}
//------------------------------------------------------------------------------
/** Get the cut rectangle's position. */
Vector2D Group::GetCutRectPosition() const
{
    return Vector2D(m_cutRect.Left, m_cutRect.Top);
}
//------------------------------------------------------------------------------
/** Get the cut rectangle's size. */
Vector2D Group::GetCutRectSize() const
{
    return Vector2D(m_cutRect.Right - m_cutRect.Left, m_cutRect.Bottom - m_cutRect.Top);
}
//------------------------------------------------------------------------------
void Group::RegisterEventAutoRemoveOnDestroy(EventHandlerBase *event)
{
    m_eventToBeDestroyedOnDelete.push_back(event);
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
