//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef PHOBJ_HPP
#define PHOBJ_HPP
//------------------------------------------------------------------------------
#include <SFML/Graphics.hpp>
//------------------------------------------------------------------------------
#include <utility>
#include <vector>
//------------------------------------------------------------------------------
#include "Object.hpp"
//------------------------------------------------------------------------------
#include <luabind/object.hpp>
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
class LinePtPt;
//------------------------------------------------------------------------------
class LinePtVect
{
public:
    LinePtVect(const Vector2D &cPt, const Vector2D &cVect);
    LinePtVect(const LinePtPt &line);
    float GetSize() const;
    float GetSqrSum() const;

public:
    Vector2D pt;
    Vector2D vect;
};
//------------------------------------------------------------------------------
class LinePtPt
{
public:
    LinePtPt(const Vector2D &cPt1, const Vector2D &cPt2);
    LinePtPt(const LinePtVect &line);
    float GetSize() const;
    float GetSqrSum() const;

public:
    Vector2D pt1;
    Vector2D pt2;
};
//------------------------------------------------------------------------------
//typedef std::pair<Vector2D, Vector2D> Line;
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A base class containing common functions to all physic objects. */
/// LUA_HEADER_COPY_OFF ///
class PhysicObject : public Object
{
public:
    PhysicObject(Group *parent, int phGroupId, const std::string &name);
    //PhysicObject(const PhysicObject& src);
    virtual ~PhysicObject();

    /// LUA_EXPORT_ON(PhysicObject, Object) ///
    /** Return true if the point \\a is inside the object, false otherwise. */
    virtual bool IsPointInside(const Vector2D &point) = 0;

    // Lua
    /** Acces the event function name (\\a fnName should be the name of an existing lua function). */
    //LUA// luaProperty(std::string, eventFunction, GetEventFunction, SetEventFunction);
    /** Set the event function name (\\a fnName should be the name of an existing lua function).\\n
      * Function prototype should be : function fnName(object1 : PhysicObject, object2 : PhysicObject, intersectionPosition : Vector2D)
      */
    void SetEventFunction(const luabind::object &luaFunction);
    /** Get the event function name. */
    luabind::object GetEventFunction() const;

    /** Get the physic group associated to the object */
    int GetPhysicGroup() const;
    /// LUA_EXPORT_OFF ///


    // No Lua
    // Lua
    //static void RegisterClassLua(lua_State *myLuaState);

    // Intersection
    virtual bool Intersect(const PhysicObject &obj, Vector2D &pos) const;

    // Event
    void OnIntersect(PhysicObject *obj1, PhysicObject *obj2, const Vector2D &pos);

    // Parent
    void ChangeParent(Group *parent);

    // Clone & Create
    virtual PhysicObject *Clone() const = 0;
    virtual PhysicObject *CloneToGroup(Group *newParent) const = 0;

    // Show
    virtual void RenderShape(sf::RenderTarget &App) const = 0;

    // Maximum interaction radius
    virtual float GetMaxInteractionRadius() const = 0;

    // Some object need to be noticed when the parent screen is changed
    virtual void OnChangeScreen(Screen *backScreen, Screen *newScreen);

    // Register a clone
    virtual void RegisterClone(Group *newParent);
protected:

    int m_group;
    luabind::object m_luaFunction;

    bool IsLineInter(const LinePtVect &l1, const LinePtVect &l2, Vector2D &pos) const;
    bool IsLineCircleInter(const LinePtVect &l1, const Vector2D &center, float radius, Vector2D &pos) const;
};
//------------------------------------------------------------------------------
class PhysicObjectLine : public PhysicObject
{
public:
    // No Lua
    PhysicObjectLine(Group *parent, int phGroupId, const std::string &name) : PhysicObject(parent, phGroupId, name) {}
    virtual bool IsPointInside(const Vector2D &point);

    virtual std::vector<Vector2D> GetAllPoints() const = 0;
    virtual void RenderShape(sf::RenderTarget &App) const;

    // Maximum interaction radius
    virtual float GetMaxInteractionRadius() const {return m_cachedMaxInterRadius;}

protected:
    float m_cachedMaxInterRadius;
    void RefreshMaxInterRadius();
};
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
#endif // PHOBJ_HPP
